#ifndef QUICKTIME_H
#define QUICKTIME_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>

// ===================== compression formats =================== //

#define QUICKTIME_RAW  "raw "
#define QUICKTIME_JPEG "jpeg"
#define QUICKTIME_YUV2 "yuv2"
#define QUICKTIME_TWOS "twos"

// ===================== private interface ===================== //

#define HEADER_LENGTH 8
#define MAXTRACKS 1024

typedef struct
{
	float values[9];
} quicktime_matrix_t;


typedef struct
{
	int version;
	long flags;
	unsigned long creation_time;
	unsigned long modification_time;
	int track_id;
	long reserved1;
	long duration;
	char reserved2[8];
	int layer;
	int alternate_group;
	float volume;
	long reserved3;
	quicktime_matrix_t matrix;
	float track_width;
	float track_height;
} quicktime_tkhd_t;


typedef struct
{
	long seed;
	long flags;
	long size;
	short int *alpha;
	short int *red;
	short int *green;
	short int *blue;
} quicktime_ctab_t;



// ===================== sample table ======================== //



// sample description

typedef struct
{
	int motion_jpeg_quantization_table;
} quicktime_mjqt_t;


typedef struct
{
	int motion_jpeg_huffman_table;
} quicktime_mjht_t;


typedef struct
{
	char format[4];
	char reserved[6];
	int data_reference;

// common to audio and video
	int version;
	int revision;
	char vendor[4];

// video description
	long temporal_quality;
	long spatial_quality;
	int width;
	int height;
	float dpi_horizontal;
	float dpi_vertical;
	long data_size;
	int frames_per_sample;
	char compressor_name[32];
	int depth;
	int ctab_id;
	quicktime_ctab_t ctab;
	float gamma;
	int fields;
	quicktime_mjqt_t mjqt;
	quicktime_mjht_t mjht;

// audio description
	int channels;
	int sample_size;
	int compression_id;
	int packet_size;
	float sample_rate;
} quicktime_stsd_table_t;


typedef struct
{
	int version;
	long flags;
	long total_entries;
	quicktime_stsd_table_t *table;
} quicktime_stsd_t;


// time to sample
typedef struct
{
	long sample_count;
	long sample_duration;
} quicktime_stts_table_t;

typedef struct
{
	int version;
	long flags;
	long total_entries;
	quicktime_stts_table_t *table;
} quicktime_stts_t;


// sync sample
typedef struct
{
	long sample;
} quicktime_stss_table_t;

typedef struct
{
	int version;
	long flags;
	long total_entries;
	quicktime_stss_table_t *table;
} quicktime_stss_t;


// sample to chunk
typedef struct
{
	long chunk;
	long samples;
	long id;
} quicktime_stsc_table_t;

typedef struct
{
	int version;
	long flags;
	long total_entries;
	
	long entries_allocated;
	quicktime_stsc_table_t *table;
} quicktime_stsc_t;


// sample size
typedef struct
{
	long size;
} quicktime_stsz_table_t;

typedef struct
{
	int version;
	long flags;
	long sample_size;
	long total_entries;

	long entries_allocated;    // used by the library for allocating a table
	quicktime_stsz_table_t *table;
} quicktime_stsz_t;


// chunk offset
typedef struct
{
	long offset;
} quicktime_stco_table_t;

typedef struct
{
	int version;
	long flags;
	long total_entries;
	
	long entries_allocated;    // used by the library for allocating a table
	quicktime_stco_table_t *table;
} quicktime_stco_t;


// sample table
typedef struct
{
	int version;
	long flags;
	quicktime_stsd_t stsd;
	quicktime_stts_t stts;
	quicktime_stss_t stss;
	quicktime_stsc_t stsc;
	quicktime_stsz_t stsz;
	quicktime_stco_t stco;
} quicktime_stbl_t;

// data reference

typedef struct
{
	long size;
	char type[4];
	int version;
	long flags;
	char *data_reference;
} quicktime_dref_table_t;

typedef struct
{
	int version;
	long flags;
	long total_entries;
	quicktime_dref_table_t *table;
} quicktime_dref_t;

// data information

typedef struct
{
	quicktime_dref_t dref;
} quicktime_dinf_t;

// video media header

typedef struct
{
	int version;
	long flags;
	int graphics_mode;
	int opcolor[3];
} quicktime_vmhd_t;


// sound media header

typedef struct
{
	int version;
	long flags;
	int balance;
	int reserved;
} quicktime_smhd_t;

// handler reference

typedef struct
{
	int version;
	long flags;
	char component_type[4];
	char component_subtype[4];
	long component_manufacturer;
	long component_flags;
	long component_flag_mask;
	char component_name[256];
} quicktime_hdlr_t;

// media information

typedef struct
{
	int is_video;
	int is_audio;
	quicktime_vmhd_t vmhd;
	quicktime_smhd_t smhd;
	quicktime_stbl_t stbl;
	quicktime_hdlr_t hdlr;
	quicktime_dinf_t dinf;
} quicktime_minf_t;


// media header

typedef struct
{
	int version;
	long flags;
	unsigned long creation_time;
	unsigned long modification_time;
	long time_scale;
	long duration;
	int language;
	int quality;
} quicktime_mdhd_t;


// media

typedef struct
{
	quicktime_mdhd_t mdhd;
	quicktime_minf_t minf;
	quicktime_hdlr_t hdlr;
} quicktime_mdia_t;

// edit list
typedef struct
{
	long duration;
	long time;
	float rate;
} quicktime_elst_table_t;

typedef struct
{
	int version;
	long flags;
	long total_entries;

	quicktime_elst_table_t *table;
} quicktime_elst_t;

typedef struct
{
	quicktime_elst_t elst;
} quicktime_edts_t;


typedef struct
{
	quicktime_tkhd_t tkhd;
	quicktime_mdia_t mdia;
	quicktime_edts_t edts;
} quicktime_trak_t;


typedef struct
{
	int version;
	long flags;
	unsigned long creation_time;
	unsigned long modification_time;
	long time_scale;
	long duration;
	float preferred_rate;
	float preferred_volume;
	char reserved[10];
	quicktime_matrix_t matrix;
	long preview_time;
	long preview_duration;
	long poster_time;
	long selection_time;
	long selection_duration;
	long current_time;
	long next_track_id;
} quicktime_mvhd_t;


typedef struct
{
	int total_tracks;

	quicktime_mvhd_t mvhd;
	quicktime_trak_t *trak[MAXTRACKS];
	quicktime_ctab_t ctab;
} quicktime_moov_t;

typedef struct
{
	long start;
	long size;
} quicktime_mdat_t;


typedef struct
{
	long start;      // byte start in file
	long end;        // byte endpoint in file
	long size;       // byte size for writing
	char type[4];
} quicktime_atom_t;

// table of pointers to every track
typedef struct
{
	quicktime_trak_t *track; // real quicktime track corresponding to this table
	int channels;            // number of audio channels in the track
	long current_position;   // current sample in output file
	long current_chunk;      // current chunk in output file
} quicktime_audio_map_t;

typedef struct
{
	quicktime_trak_t *track;
	long current_position;
	long current_chunk;
} quicktime_video_map_t;

// file descriptor passed to all routines

typedef struct
{
	FILE *stream;
	long total_length;
	quicktime_mdat_t mdat;
	quicktime_moov_t moov;
	int rd;
	int wr;

// mapping of audio channels to movie tracks
// one audio map entry exists for each channel
	int total_atracks;
	quicktime_audio_map_t *atracks;

// mapping of video tracks to movie tracks
	int total_vtracks;
	quicktime_video_map_t *vtracks;

// for begining and ending frame writes where the user wants to write the 
// file descriptor directly
	long offset;
} quicktime_t;

// atom handling routines
int quicktime_atom_reset(quicktime_atom_t *atom);
int quicktime_atom_read_header(FILE *stream, quicktime_atom_t *atom);
int quicktime_atom_is(quicktime_atom_t *atom, char *type);
long quicktime_atom_read_size(char *data);
int quicktime_atom_read_type(char *data, char *output);
int quicktime_atom_skip(quicktime_t *file, quicktime_atom_t *atom);
int quicktime_atom_write_header(quicktime_t *file, quicktime_atom_t *atom, char *text);
int quicktime_atom_write_footer(quicktime_t *file, quicktime_atom_t *atom);

// utilities for reading data types
int quicktime_read_data(quicktime_t *file, char *data, int size);
int quicktime_write_data(quicktime_t *file, char *data, int size);
int quicktime_read_pascal(quicktime_t *file, char *data);
int quicktime_write_pascal(quicktime_t *file, char *data);
float quicktime_read_fixed32(quicktime_t *file);
int quicktime_write_fixed32(quicktime_t *file, float number);
float quicktime_read_fixed16(quicktime_t *file);
int quicktime_write_fixed16(quicktime_t *file, float number);
long quicktime_read_int32(quicktime_t *file);
int quicktime_write_int32(quicktime_t *file, long number);
long quicktime_read_int24(quicktime_t *file);
int quicktime_write_int24(quicktime_t *file, long number);
int quicktime_read_int16(quicktime_t *file);
int quicktime_write_int16(quicktime_t *file, int number);
int quicktime_read_char(quicktime_t *file);
int quicktime_write_char(quicktime_t *file, char x);
int quicktime_read_char32(quicktime_t *file, char *string);
int quicktime_write_char32(quicktime_t *file, char *string);
int quicktime_copy_char32(char *output, char *input);
long quicktime_position(quicktime_t *file);
int quicktime_set_position(quicktime_t *file, long position);
int quicktime_print_chars(char *desc, char *input, int len);

// initializers for every atom
int quicktime_matrix_init(quicktime_matrix_t *matrix);
int quicktime_edts_init_table(quicktime_edts_t *edts);
int quicktime_edts_init(quicktime_edts_t *edts);
int quicktime_elst_init(quicktime_elst_t *elst);
int quicktime_elst_init_all(quicktime_elst_t *elst);
int quicktime_elst_table_init(quicktime_elst_table_t *table); // initialize a table
int quicktime_tkhd_init(quicktime_tkhd_t *tkhd);
int quicktime_tkhd_init_video(quicktime_t *file, quicktime_tkhd_t *tkhd, int frame_w, int frame_h);
int quicktime_ctab_init(quicktime_ctab_t *ctab);
int quicktime_mjqt_init(quicktime_mjqt_t *mjqt);
int quicktime_mjht_init(quicktime_mjht_t *mjht);
int quicktime_stsd_table_init(quicktime_stsd_table_t *table);
int quicktime_stsd_init(quicktime_stsd_t *stsd);
int quicktime_stsd_init_table(quicktime_stsd_t *stsd);
int quicktime_stsd_init_video(quicktime_t *file, quicktime_stsd_t *stsd, int frame_w, int frame_h, float frame_rate, char *compression);
int quicktime_stsd_init_audio(quicktime_t *file, quicktime_stsd_t *stsd, int channels, int sample_rate, int bits, char *compressor);
int quicktime_stts_init(quicktime_stts_t *stts);
int quicktime_stts_init_table(quicktime_stts_t *stts);
int quicktime_stts_init_video(quicktime_t *file, quicktime_stts_t *stts, float frame_rate);
int quicktime_stts_init_audio(quicktime_t *file, quicktime_stts_t *stts, int sample_rate);
int quicktime_stss_init(quicktime_stss_t *stss);
int quicktime_stsc_init(quicktime_stsc_t *stsc);
int quicktime_stsc_init_video(quicktime_t *file, quicktime_stsc_t *stsc);
int quicktime_stsc_init_audio(quicktime_t *file, quicktime_stsc_t *stsc, int sample_rate);
int quicktime_stsz_init(quicktime_stsz_t *stsz);
int quicktime_stsz_init_video(quicktime_t *file, quicktime_stsz_t *stsz);
int quicktime_stsz_init_audio(quicktime_t *file, quicktime_stsz_t *stsz, int channels, int bits);
int quicktime_stco_init(quicktime_stco_t *stco);
int quicktime_stco_init_common(quicktime_t *file, quicktime_stco_t *stco);
int quicktime_stbl_init(quicktime_stbl_t *tkhd);
int quicktime_stbl_init_video(quicktime_t *file, quicktime_stbl_t *stbl, int frame_w, int frame_h, float frame_rate, char *compressor);
int quicktime_stbl_init_audio(quicktime_t *file, quicktime_stbl_t *stbl, int channels, int sample_rate, int bits, char *compressor);
int quicktime_vmhd_init(quicktime_vmhd_t *vmhd);
int quicktime_vmhd_init_video(quicktime_t *file, quicktime_vmhd_t *vmhd, int frame_w, int frame_h, float frame_rate);
int quicktime_smhd_init(quicktime_smhd_t *smhd);
int quicktime_dref_table_init(quicktime_dref_table_t *table);
int quicktime_dref_init_all(quicktime_dref_t *dref);
int quicktime_dref_init(quicktime_dref_t *dref);
int quicktime_dinf_init_all(quicktime_dinf_t *dinf);
int quicktime_dinf_init(quicktime_dinf_t *dinf);
int quicktime_minf_init(quicktime_minf_t *minf);
int quicktime_minf_init_video(quicktime_t *file, quicktime_minf_t *minf, int frame_w, int frame_h, float frame_rate, char *compressor);
int quicktime_minf_init_audio(quicktime_t *file, quicktime_minf_t *minf, int channels, int sample_rate, int bits, char *compressor);
int quicktime_mdhd_init(quicktime_mdhd_t *mdhd);
int quicktime_mdhd_init_video(quicktime_t *file, quicktime_mdhd_t *mdhd, int frame_w, int frame_h, float frame_rate);
int quicktime_mdhd_init_audio(quicktime_t *file, quicktime_mdhd_t *mdhd, int channels, int sample_rate, int bits, char *compressor);
int quicktime_mdia_init(quicktime_mdia_t *mdia);
int quicktime_mdia_init_video(quicktime_t *file, quicktime_mdia_t *mdia, int frame_w, int frame_h, float frame_rate, char *compressor);
int quicktime_mdia_init_audio(quicktime_t *file, quicktime_mdia_t *mdia, int channels, int sample_rate, int bits, char *compressor);
int quicktime_trak_init(quicktime_trak_t *trak);
int quicktime_trak_init_video(quicktime_t *file, quicktime_trak_t *trak, int frame_w, int frame_h, float frame_rate, char *compressor);
int quicktime_trak_init_audio(quicktime_t *file, quicktime_trak_t *trak, int channels, int sample_rate, int bits, char *compressor);
int quicktime_mvhd_init(quicktime_mvhd_t *mvhd);
int quicktime_mhvd_init_video(quicktime_t *file, quicktime_mvhd_t *mvhd, float frame_rate);
int quicktime_moov_init(quicktime_moov_t *moov);
int quicktime_mdat_init(quicktime_mdat_t *mdat);
int quicktime_init(quicktime_t *file);
int quicktime_hdlr_init(quicktime_hdlr_t *hdlr);
int quicktime_hdlr_init_video(quicktime_hdlr_t *hdlr);
int quicktime_hdlr_init_audio(quicktime_hdlr_t *hdlr);
int quicktime_hdlr_init_data(quicktime_hdlr_t *hdlr);

// deleters for every atom
int quicktime_matrix_delete(quicktime_matrix_t *matrix);
int quicktime_edts_delete(quicktime_edts_t *edts);
int quicktime_elst_delete(quicktime_elst_t *elst);
int quicktime_elst_table_delete(quicktime_elst_table_t *table);
int quicktime_tkhd_delete(quicktime_tkhd_t *tkhd);
int quicktime_ctab_delete(quicktime_ctab_t *ctab);
int quicktime_mjqt_delete(quicktime_mjqt_t *mjqt);
int quicktime_mjht_delete(quicktime_mjht_t *mjht);
int quicktime_stsd_table_delete(quicktime_stsd_table_t *table);
int quicktime_stsd_delete(quicktime_stsd_t *stsd);
int quicktime_stts_delete(quicktime_stts_t *stts);
int quicktime_stss_delete(quicktime_stss_t *stss);
int quicktime_stsc_delete(quicktime_stsc_t *stsc);
int quicktime_stsz_delete(quicktime_stsz_t *stsz);
int quicktime_stco_delete(quicktime_stco_t *stco);
int quicktime_stbl_delete(quicktime_stbl_t *stbl);
int quicktime_vmhd_delete(quicktime_vmhd_t *vmhd);
int quicktime_smhd_delete(quicktime_smhd_t *smhd);
int quicktime_dref_table_delete(quicktime_dref_table_t *table);
int quicktime_dref_delete(quicktime_dref_t *dref);
int quicktime_dinf_delete(quicktime_dinf_t *dinf);
int quicktime_minf_delete(quicktime_minf_t *minf);
int quicktime_mdhd_delete(quicktime_mdhd_t *mdhd);
int quicktime_mdia_delete(quicktime_mdia_t *mdia);
int quicktime_trak_delete(quicktime_trak_t *trak);
quicktime_trak_t* quicktime_add_track(quicktime_moov_t *moov);
int quicktime_mvhd_delete(quicktime_mvhd_t *mvhd);
int quicktime_moov_delete(quicktime_moov_t *moov);
int quicktime_mdat_delete(quicktime_mdat_t *moov);
int quicktime_delete(quicktime_t *file);
int quicktime_hdlr_delete(quicktime_hdlr_t *hdlr);


// dumpers for every atom
int quicktime_matrix_dump(quicktime_matrix_t *matrix);
int quicktime_edts_dump(quicktime_edts_t *edts);
int quicktime_elst_dump(quicktime_elst_t *elst);
int quicktime_elst_table_dump(quicktime_elst_table_t *table);
int quicktime_tkhd_dump(quicktime_tkhd_t *tkhd);
int quicktime_ctab_dump(quicktime_ctab_t *ctab);
int quicktime_mjqt_dump(quicktime_mjqt_t *mjqt);
int quicktime_mjht_dump(quicktime_mjht_t *mjht);
int quicktime_stsd_table_dump(void *minf_ptr, quicktime_stsd_table_t *table);
int quicktime_stsd_audio_dump(quicktime_stsd_table_t *table);
int quicktime_stsd_video_dump(quicktime_stsd_table_t *table);
int quicktime_stsd_dump(void *minf_ptr, quicktime_stsd_t *stsd);
int quicktime_stts_dump(quicktime_stts_t *stts);
int quicktime_stss_dump(quicktime_stss_t *stss);
int quicktime_stsc_dump(quicktime_stsc_t *stsc);
int quicktime_stsz_dump(quicktime_stsz_t *stsz);
int quicktime_stco_dump(quicktime_stco_t *stco);
int quicktime_stbl_dump(void *minf_ptr, quicktime_stbl_t *stbl);
int quicktime_vmhd_dump(quicktime_vmhd_t *vmhd);
int quicktime_smhd_dump(quicktime_smhd_t *smhd);
int quicktime_dref_table_dump(quicktime_dref_table_t *table);
int quicktime_dref_dump(quicktime_dref_t *dref);
int quicktime_dinf_dump(quicktime_dinf_t *dinf);
int quicktime_minf_dump(quicktime_minf_t *minf);
int quicktime_mdhd_dump(quicktime_mdhd_t *mdhd);
int quicktime_mdia_dump(quicktime_mdia_t *mdia);
int quicktime_trak_dump(quicktime_trak_t *trak);
int quicktime_dump_mvhd(quicktime_mvhd_t *mvhd);
int quicktime_moov_dump(quicktime_moov_t *moov);
int quicktime_hdlr_dump(quicktime_hdlr_t *hdlr);

// storage for every type of atom
int quicktime_read_moov(quicktime_t *file, quicktime_moov_t *moov, quicktime_atom_t *parent_atom);
int quicktime_write_moov(quicktime_t *file, quicktime_moov_t *moov);
int quicktime_read_mdat(quicktime_t *file, quicktime_mdat_t *mdat, quicktime_atom_t *parent_atom);
int quicktime_write_mdat(quicktime_t *file, quicktime_mdat_t *mdat);
int quicktime_read_matrix(quicktime_t *file, quicktime_matrix_t *matrix);
int quicktime_write_matrix(quicktime_t *file, quicktime_matrix_t *matrix);
int quicktime_read_ctab(quicktime_t *file, quicktime_ctab_t *ctab);
int quicktime_read_trak(quicktime_t *file, quicktime_trak_t *trak, quicktime_atom_t *parent_atom);
int quicktime_write_trak(quicktime_t *file, quicktime_trak_t *trak, long moov_time_scale);
int quicktime_read_edts(quicktime_t *file, quicktime_edts_t *edts, quicktime_atom_t *edts_atom);
int quicktime_write_edts(quicktime_t *file, quicktime_edts_t *edts, long duration);
int quicktime_read_elst(quicktime_t *file, quicktime_elst_t *elst);
int quicktime_write_elst(quicktime_t *file, quicktime_elst_t *elst, long duration);
int quicktime_read_elst_table(quicktime_t *file, quicktime_elst_table_t *table);
int quicktime_write_elst_table(quicktime_t *file, quicktime_elst_table_t *table, long duration);
int quicktime_read_tkhd(quicktime_t *file, quicktime_tkhd_t *tkhd);
int quicktime_write_tkhd(quicktime_t *file, quicktime_tkhd_t *tkhd);
int quicktime_read_mvhd(quicktime_t *file, quicktime_mvhd_t *mvhd);
int quicktime_write_mvhd(quicktime_t *file, quicktime_mvhd_t *mvhd);
int quicktime_read_mdhd(quicktime_t *file, quicktime_mdhd_t *mdhd);
int quicktime_read_mdia(quicktime_t *file, quicktime_mdia_t *mdia, quicktime_atom_t *parent_atom);
int quicktime_write_mdia(quicktime_t *file, quicktime_mdia_t *mdia);
int quicktime_read_dref_table(quicktime_t *file, quicktime_dref_table_t *table);
int quicktime_write_dref_table(quicktime_t *file, quicktime_dref_table_t *table);
int quicktime_read_dref(quicktime_t *file, quicktime_dref_t *dref);
int quicktime_write_dref(quicktime_t *file, quicktime_dref_t *dref);
int quicktime_read_dinf(quicktime_t *file, quicktime_dinf_t *dinf, quicktime_atom_t *dinf_atom);
int quicktime_write_dinf(quicktime_t *file, quicktime_dinf_t *dinf);
int quicktime_read_minf(quicktime_t *file, quicktime_minf_t *minf, quicktime_atom_t *parent_atom);
int quicktime_read_vmhd(quicktime_t *file, quicktime_vmhd_t *vmhd);
int quicktime_write_vmhd(quicktime_t *file, quicktime_vmhd_t *vmhd);
int quicktime_read_smhd(quicktime_t *file, quicktime_smhd_t *smhd);
int quicktime_write_smhd(quicktime_t *file, quicktime_smhd_t *smhd);
int quicktime_read_stbl(quicktime_t *file, quicktime_minf_t *minf, quicktime_stbl_t *stbl, quicktime_atom_t *parent_atom);
int quicktime_write_stbl(quicktime_t *file, quicktime_minf_t *minf, quicktime_stbl_t *stbl);
int quicktime_read_stsd(quicktime_t *file, quicktime_minf_t *minf, quicktime_stsd_t *stsd);
int quicktime_read_stsd_table(quicktime_t *file, quicktime_minf_t *minf, quicktime_stsd_table_t *table);
int quicktime_write_stsd_table(quicktime_t *file, quicktime_minf_t *minf, quicktime_stsd_table_t *table);
int quicktime_read_stsd_audio(quicktime_t *file, quicktime_stsd_table_t *table, quicktime_atom_t *parent_atom);
int quicktime_write_stsd_audio(quicktime_t *file, quicktime_stsd_table_t *table);
int quicktime_read_stsd_video(quicktime_t *file, quicktime_stsd_table_t *table, quicktime_atom_t *parent_atom);
int quicktime_write_stsd_video(quicktime_t *file, quicktime_stsd_table_t *table);
int quicktime_read_stts(quicktime_t *file, quicktime_stts_t *stts);
int quicktime_write_stts(quicktime_t *file, quicktime_stts_t *stts);
int quicktime_read_stss(quicktime_t *file, quicktime_stss_t *stss);
int quicktime_write_stss(quicktime_t *file, quicktime_stss_t *stss);
int quicktime_read_stsc(quicktime_t *file, quicktime_stsc_t *stsc);
int quicktime_write_stsc(quicktime_t *file, quicktime_stsc_t *stsc);
int quicktime_read_stsz(quicktime_t *file, quicktime_stsz_t *stsz);
int quicktime_write_stsz(quicktime_t *file, quicktime_stsz_t *stsz);
int quicktime_read_stco(quicktime_t *file, quicktime_stco_t *stco);
int quicktime_write_stco(quicktime_t *file, quicktime_stco_t *stco);
int quicktime_read_hdlr(quicktime_t *file, quicktime_hdlr_t *hdlr);
int quicktime_write_hdlr(quicktime_t *file, quicktime_hdlr_t *hdlr);

// chunks always start on 1
// samples start on 0

// queries for every atom
// the starting sample in the given chunk
long quicktime_sample_of_chunk(quicktime_trak_t *trak, long chunk);

// number of samples in the chunk
long quicktime_chunk_samples(quicktime_trak_t *trak, long chunk);

// the chunk containing the sample
int quicktime_chunk_of_sample(long *chunk_sample, long *chunk, quicktime_trak_t *trak, long sample);

// the byte offset from mdat start of the chunk
long quicktime_chunk_to_offset(quicktime_trak_t *trak, long chunk);

// the chunk of any offset from mdat start
long quicktime_offset_to_chunk(long *chunk_offset, quicktime_trak_t *trak, long offset);

// the total number of samples in the track depending on the access mode
long quicktime_track_samples(quicktime_t *file, quicktime_trak_t *trak);

// length of track in timescale
int quicktime_trak_duration(quicktime_trak_t *trak, long *duration, long *timescale);

// total bytes between the two samples
long quicktime_sample_range_size(quicktime_trak_t *trak, long chunk_sample, long sample);

// update the position pointers in all the tracks after a set_position
int quicktime_update_positions(quicktime_t *file);

// converting between mdat offsets to samples
long quicktime_sample_to_offset(quicktime_trak_t *trak, long sample);
long quicktime_offset_to_sample(quicktime_trak_t *trak, long offset);

quicktime_trak_t* quicktime_add_trak(quicktime_moov_t *moov);
int quicktime_delete_trak(quicktime_moov_t *moov);
int quicktime_get_timescale(float frame_rate);

// update all the tables after writing a buffer
// set sample_size to 0 if no sample size should be set
int quicktime_update_tables(quicktime_trak_t *trak, 
							long offset, 
							long chunk, 
							long sample, 
							long samples, 
							long sample_size);
int quicktime_update_stco(quicktime_stco_t *stco, long chunk, long offset);
int quicktime_update_stsc(quicktime_stsc_t *stsc, long chunk, long samples);
int quicktime_update_stsz(quicktime_stsz_t *stsz, long sample, long sample_size);
int quicktime_trak_fix_counts(quicktime_t *file, quicktime_trak_t *trak);
int quicktime_update_durations(quicktime_moov_t *moov);
unsigned long quicktime_current_time();

// for making streamable
int quicktime_shift_offsets(quicktime_moov_t *moov, long offset);












// =========================== public interface ========================= //

// return 1 if the file is a quicktime file
int quicktime_check_sig(char *path);

// make the quicktime file streamable
int quicktime_make_streamable(char *path);

// call this first to open the file and create all the objects
quicktime_t* quicktime_open(char *filename, int rd, int wr);

// Read all the information about the file.
// Requires a MOOV atom be present in the file.
// If no MOOV atom exists return 1 else return 0.
int quicktime_read_info(quicktime_t *file);

// set up tracks in a new file after opening
// returns the number of quicktime tracks allocated
// audio is stored two channels per quicktime track
int quicktime_set_audio(quicktime_t *file, int channels, long sample_rate, int bits, char *compressor);
// video is stored one layer per quicktime track
int quicktime_set_video(quicktime_t *file, int tracks, int frame_w, int frame_h, float frame_rate, char *compressor);

// close the file and delete all the objects
int quicktime_close(quicktime_t *file);

// get length information
// channel numbers start on 1 for audio and video
long quicktime_audio_length(quicktime_t *file, int track);
long quicktime_video_length(quicktime_t *file, int track);

// get position information
long quicktime_audio_position(quicktime_t *file, int track);
long quicktime_video_position(quicktime_t *file, int track);

// get file information
int quicktime_video_tracks(quicktime_t *file);
int quicktime_audio_tracks(quicktime_t *file);

int quicktime_has_audio(quicktime_t *file);
long quicktime_sample_rate(quicktime_t *file, int track);
int quicktime_audio_bits(quicktime_t *file, int track);
int quicktime_track_channels(quicktime_t *file, int track);
char* quicktime_audio_compressor(quicktime_t *file, int track);

int quicktime_has_video(quicktime_t *file);
int quicktime_video_width(quicktime_t *file, int track);
int quicktime_video_height(quicktime_t *file, int track);
float quicktime_frame_rate(quicktime_t *file, int track);
char* quicktime_video_compressor(quicktime_t *file, int track);
long quicktime_frame_size(quicktime_t *file, long frame, int track);

// get the quicktime track and channel that the audio channel belongs to
// channels and tracks start on 0
int quicktime_channel_location(quicktime_t *file, int *quicktime_track, int *quicktime_channel, int channel);

// file positioning
int quicktime_seek_end(quicktime_t *file);
int quicktime_seek_start(quicktime_t *file);

// set position of file descriptor relative to a track
int quicktime_set_audio_position(quicktime_t *file, long sample, int track);
int quicktime_set_video_position(quicktime_t *file, long frame, int track);

// write data for one quicktime track
// the user must handle conversion to the channels in this track
int quicktime_write_audio(quicktime_t *file, char *audio_buffer, long samples, int track);
int quicktime_write_frame(quicktime_t *file, unsigned char *video_buffer, long bytes, int track);

// for writing a frame using a library that needs a file descriptor
int quicktime_write_frame_init(quicktime_t *file, int track); // call before fwrite
FILE* quicktime_get_fd(quicktime_t *file);     // return a file descriptor
int quicktime_write_frame_end(quicktime_t *file, int track); // call after fwrite

// read data
long quicktime_read_audio(quicktime_t *file, char *audio_buffer, long samples, int track);
long quicktime_read_frame(quicktime_t *file, unsigned char *video_buffer, int track);

// for reading frame using a library that needs a file descriptor
int quicktime_read_frame_init(quicktime_t *file, int track);
int quicktime_read_frame_end(quicktime_t *file, int track);

// dump the file structures
int quicktime_dump(quicktime_t *file);

#ifdef __cplusplus
}
#endif

#endif
